@@ -9,7 +9,7 @@
 #  (      _ \     /  |     (   | (_ |    |      |
 # \___| _/  _\ _|_\ ____| \___/ \___|   _|     _|
 
-# secrets_manager.py
+# src/carlogtt_python_library/aws_boto3/secrets_manager.py
 # Created 11/22/23 - 12:25 PM UK Time (London) by carlogtt
 # Copyright (c) Amazon.com Inc. All Rights Reserved.
 # AMAZON.COM CONFIDENTIAL
@@ -34,21 +34,16 @@ This module ...
 # Standard Library Imports
 import json
 import logging
-from typing import Any, Literal, Optional
+from typing import Any, Optional
 
 # Third Party Library Imports
-import boto3
 import botocore.exceptions
-from mypy_boto3_secretsmanager.client import SecretsManagerClient
-from mypy_boto3_secretsmanager.type_defs import (
-    DeleteSecretRequestTypeDef,
-    DeleteSecretResponseTypeDef,
-    ListSecretsRequestTypeDef,
-    SecretListEntryTypeDef,
-)
+import mypy_boto3_secretsmanager
+from mypy_boto3_secretsmanager import type_defs
 
 # Local Folder (Relative) Imports
 from .. import exceptions
+from . import aws_service_base
 
 # END IMPORTS
 # ======================================================================
@@ -63,10 +58,10 @@ __all__ = [
 module_logger = logging.getLogger(__name__)
 
 # Type aliases
-#
+SecretsManagerClient = mypy_boto3_secretsmanager.client.SecretsManagerClient
 
 
-class SecretsManager:
+class SecretsManager(aws_service_base.AwsServiceBase[SecretsManagerClient]):
     """
     The SecretsManager class provides a simplified interface for
     interacting with Amazon SecretsManager services within a Python
@@ -114,84 +109,19 @@ class SecretsManager:
         caching: bool = False,
         client_parameters: Optional[dict[str, Any]] = None,
     ) -> None:
-        self._aws_region_name = aws_region_name
-        self._aws_profile_name = aws_profile_name
-        self._aws_access_key_id = aws_access_key_id
-        self._aws_secret_access_key = aws_secret_access_key
-        self._aws_session_token = aws_session_token
-        self._caching = caching
-        self._cache: dict[str, Any] = dict()
-        self._aws_service_name: Literal['secretsmanager'] = "secretsmanager"
-        self._client_parameters = client_parameters if client_parameters else dict()
+        super().__init__(
+            aws_region_name=aws_region_name,
+            aws_profile_name=aws_profile_name,
+            aws_access_key_id=aws_access_key_id,
+            aws_secret_access_key=aws_secret_access_key,
+            aws_session_token=aws_session_token,
+            caching=caching,
+            client_parameters=client_parameters,
+            aws_service_name="secretsmanager",
+            exception_type=exceptions.SecretsManagerError,
+        )
 
-    @property
-    def _client(self) -> SecretsManagerClient:
-        """
-        Returns a SecretsManger client.
-        Caches the client if caching is enabled.
-
-        :return: The SecretsManagerClient.
-        """
-
-        if self._caching:
-            if self._cache.get('client') is None:
-                self._cache['client'] = self._get_boto_client()
-            return self._cache['client']
-
-        else:
-            return self._get_boto_client()
-
-    def _get_boto_client(self) -> SecretsManagerClient:
-        """
-        Create a low-level SecretsManager client.
-
-        :return: The SecretsManagerClient.
-        :raise SecretsManagerError: If operation fails.
-        """
-
-        try:
-            boto_session = boto3.session.Session(
-                region_name=self._aws_region_name,
-                profile_name=self._aws_profile_name,
-                aws_access_key_id=self._aws_access_key_id,
-                aws_secret_access_key=self._aws_secret_access_key,
-                aws_session_token=self._aws_session_token,
-            )
-            client = boto_session.client(
-                service_name=self._aws_service_name, **self._client_parameters
-            )
-
-            return client
-
-        except botocore.exceptions.ClientError as ex:
-            raise exceptions.SecretsManagerError(f"Operation failed! - {str(ex.response)}")
-
-        except Exception as ex:
-            raise exceptions.SecretsManagerError(f"Operation failed! - {str(ex)}")
-
-    def invalidate_client_cache(self) -> None:
-        """
-        Clears the cached client, if caching is enabled.
-
-        This method allows manually invalidating the cached client,
-        forcing a new client instance to be created on the next access.
-        Useful if AWS credentials have changed or if there's a need to
-        connect to a different region within the same instance
-        lifecycle.
-
-        :return: None.
-        :raise SecretsManagerError: Raises an error if caching is not
-               enabled for this instance.
-        """
-
-        if not self._cache:
-            raise exceptions.SecretsManagerError(
-                f"Session caching is not enabled for this instance of {self.__class__.__qualname__}"
-            )
-
-        self._cache['client'] = None
-
-    def get_all_secrets(self) -> list[SecretListEntryTypeDef]:
+    def get_all_secrets(self) -> list[type_defs.SecretListEntryTypeDef]:
         """
         Retrieves a list of all secrets stored in AWS Secrets Manager.
 
@@ -206,7 +136,7 @@ class SecretsManager:
         :raise SecretsManagerError: If operation fails.
         """
 
-        list_secrets_args: ListSecretsRequestTypeDef = {}
+        list_secrets_args: type_defs.ListSecretsRequestTypeDef = {}
         secrets = []
 
         try:
@@ -215,7 +145,7 @@ class SecretsManager:
                     secretsmanager_response = self._client.list_secrets(**list_secrets_args)
 
                 except botocore.exceptions.ClientError as ex_inner:
-                    raise exceptions.SecretsManagerError(f"{str(ex_inner.response)}")
+                    raise exceptions.SecretsManagerError(str(ex_inner.response))
 
                 secrets.extend(secretsmanager_response["SecretList"])
 
@@ -227,7 +157,7 @@ class SecretsManager:
             return secrets
 
         except Exception as ex:
-            raise exceptions.SecretsManagerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SecretsManagerError(str(ex)) from None
 
     def get_secret(self, secret_id: str, **kwargs) -> Optional[dict[str, str]]:
         """
@@ -246,14 +176,17 @@ class SecretsManager:
         :raise SecretsManagerError: If operation fails.
         """
 
+        get_secret_value_payload: type_defs.GetSecretValueRequestTypeDef = {
+            'SecretId': secret_id,
+            **kwargs,  # type: ignore
+        }
+
         try:
             try:
-                secretsmanager_response = self._client.get_secret_value(
-                    SecretId=secret_id, **kwargs
-                )
+                secretsmanager_response = self._client.get_secret_value(**get_secret_value_payload)
 
             except botocore.exceptions.ClientError as ex_inner:
-                raise exceptions.SecretsManagerError(f"{str(ex_inner.response)}")
+                raise exceptions.SecretsManagerError(str(ex_inner.response))
 
             secret_str = secretsmanager_response.get('SecretString')
 
@@ -267,7 +200,7 @@ class SecretsManager:
             return secret
 
         except Exception as ex:
-            raise exceptions.SecretsManagerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SecretsManagerError(str(ex)) from None
 
     def get_secret_password(self, secret_id: str, **kwargs) -> str:
         """
@@ -298,7 +231,7 @@ class SecretsManager:
         secret_id: str,
         recovery_days: int = 30,
         force_delete: bool = False,
-    ) -> DeleteSecretResponseTypeDef:
+    ) -> type_defs.DeleteSecretResponseTypeDef:
         """
         Deletes a secret from AWS Secrets Manager.
         This method supports both immediate deletion and scheduled
@@ -320,7 +253,9 @@ class SecretsManager:
         :raise SecretsManagerError: If operation fails.
         """
 
-        delete_secret_args: DeleteSecretRequestTypeDef = {'SecretId': secret_id}
+        delete_secret_args: type_defs.DeleteSecretRequestTypeDef = {
+            'SecretId': secret_id,
+        }
 
         if force_delete:
             delete_secret_args['ForceDeleteWithoutRecovery'] = True
@@ -334,7 +269,7 @@ class SecretsManager:
             return secretsmanager_response
 
         except botocore.exceptions.ClientError as ex:
-            raise exceptions.SecretsManagerError(f"Operation failed! - {str(ex.response)}")
+            raise exceptions.SecretsManagerError(str(ex.response)) from None
 
         except Exception as ex:
-            raise exceptions.SecretsManagerError(f"Operation failed! - {str(ex)}")
+            raise exceptions.SecretsManagerError(str(ex)) from None
